
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <math.h>


#define INDEX    unsigned int
#define VALUE    float


////////////////////////////////////////////////////////////
// data structures
//


// sparse vector entry
struct Entry
{
	INDEX index;
	VALUE value;
};
#define  endOfVector  ((INDEX)-1)


// data set
struct Data
{
	INDEX num;                   // number of features
	INDEX ell;                   // number of points
	struct Entry** data;         // array of sparse vectors, one per feature
	struct Entry* storage;       // linear memory
	VALUE* label;                // dense label vector, one entry per point
};


struct Hypothesis
{
	INDEX num;                   // number of features
	struct Entry* weight;        // sparse weight vector
	VALUE bias;                  // bias term (currently unused)
};


////////////////////////////////////////////////////////////
// load data from file
//

#define MAXLINELENGTH 1048576

// return 0 on success and an error code otherwise:
// 1: file open failed
// 2: file format violation
// 3: line too long (more than MAXLINELENGTH characters)
// 4: out of memory
// 5: unexpected label value (should be exactly two different values)
int load(const char* filename, struct Data* data)
{
	printf("loading file %s: ...", filename); fflush(stdout);

	FILE* file = fopen(filename, "r");
	if (file == NULL) return 1;

	char buffer[MAXLINELENGTH + 1];
	buffer[MAXLINELENGTH] = 0;
	char* pos;
	char* end;

	int mode = 1;

	// pass 1: find number of points, number of features, and number of entries
	INDEX num_points = 0;
	INDEX num_features = 0;
	size_t num_entries = 0;
	int has_zero = 0;
	pos = buffer + MAXLINELENGTH;
	VALUE label[2];
	unsigned int labels = 0;
	while (1)
	{
		// make sure we have a full line in the buffer
		const char* p;
		for (p=pos; (*p != 0) && (*p != '\n'); p++);
		if (*p == 0)
		{
			INDEX size = (buffer + MAXLINELENGTH - pos);
			memcpy(buffer, pos, size);
			pos = buffer;
			INDEX n = fread(buffer + size, 1, MAXLINELENGTH - size, file);
			if (n == 0) break;
			buffer[size + n] = 0;
			for (p=pos; (*p != 0) && (*p != '\n'); p++);
			if (*p == 0) { fclose(file); return 3; }
		}

		// ignore line with white space only
		while (*pos == ' ' || *pos == '\t') pos++;
		if (*pos == '\n')
		{
			pos++;
			continue;
		}

		// parse and record labels
		VALUE l = strtod(pos, &end);
		if (end == pos) { fclose(file); return 2; }
		pos = end;
		if (mode == 1)
		{
			unsigned int ll = 0;
			for (; ll<labels; ll++) if (label[ll] == l) break;
			if (ll == labels)
			{
				if (ll == 2) mode = 2;
				else
				{
					label[ll] = l;
					labels++;
				}
			}
		}

		// parse features
		while (1)
		{
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos == '\n')
			{
				pos++;
				break;
			}

			INDEX index = strtoll(pos, &end, 10);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			if (index >= num_features) num_features = index + 1;
			if (index == 0) has_zero = 1;
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos != ':') { fclose(file); return 2; }
			else pos++;
			while (*pos == ' ' || *pos == '\t') pos++;
			strtod(pos, &end);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			num_entries++;
		}

		num_points++;
	}
	rewind(file);
	printf("."); fflush(stdout);

	// check if labels are for binary classification
	if (mode == 1)
	{
		if (labels != 2) mode = 2;
		else if (label[0] > label[1])
		{
			VALUE tmp = label[0];
			label[0] = label[1];
			label[1] = tmp;
		}
	}

	if (! has_zero) num_features--;

	// prepare memory
	data->num = num_features;
	data->ell = num_points;
	data->label = (VALUE*)(void*)malloc(num_points * sizeof(VALUE));
	data->data = (struct Entry**)(void*)malloc((num_features + 1) * sizeof(struct Entry*));
	data->storage = (struct Entry*)(void*)malloc((num_entries) * sizeof(struct Entry));
	INDEX* feature = (INDEX*)(void*)calloc(num_features, sizeof(INDEX));
	if (data->label == NULL || data->data == NULL || data->storage == NULL || feature == NULL)
	{
		free(data->label);
		free(data->data);
		free(data->storage);
		free(feature);
		return 4;
	}

	// pass 2: count entries per feature
	pos = buffer + MAXLINELENGTH;
	while (1)
	{
		// make sure we have a full line in the buffer
		const char* p;
		for (p=pos; (*p != 0) && (*p != '\n'); p++);
		if (*p == 0)
		{
			INDEX size = (buffer + MAXLINELENGTH - pos);
			memcpy(buffer, pos, size);
			pos = buffer;
			INDEX n = fread(buffer + size, 1, MAXLINELENGTH - size, file);
			if (n == 0) break;
			buffer[size + n] = 0;
			for (p=pos; (*p != 0) && (*p != '\n'); p++);
			if (*p == 0) { fclose(file); return 3; }
		}

		// ignore line with white space only
		while (*pos == ' ' || *pos == '\t') pos++;
		if (*pos == '\n')
		{
			pos++;
			continue;
		}

		// parse label
		strtod(pos, &end);
		if (end == pos) { fclose(file); return 2; }
		pos = end;

		// parse features
		while (1)
		{
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos == '\n')
			{
				pos++;
				break;
			}

			INDEX index = strtoll(pos, &end, 10);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			if (! has_zero) index--;
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos != ':') { fclose(file); return 2; }
			else pos++;
			while (*pos == ' ' || *pos == '\t') pos++;
			strtod(pos, &end);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			feature[index]++;
		}
	}
	rewind(file);
	printf("."); fflush(stdout);

	// prepare storage and start indices
	INDEX spos = 0;
	for (INDEX i=0; i<num_features; i++)
	{
		data->data[i] = data->storage + spos;
		spos += feature[i];
		feature[i] = 0;
	}
	data->data[num_features] = data->storage + spos;   // end pointer

	// pass 3: read data
	pos = buffer + MAXLINELENGTH;
	INDEX point = 0;
	while (1)
	{
		// make sure we have a full line in the buffer
		const char* p;
		for (p=pos; (*p != 0) && (*p != '\n'); p++);
		if (*p == 0)
		{
			INDEX size = (buffer + MAXLINELENGTH - pos);
			memcpy(buffer, pos, size);
			pos = buffer;
			INDEX n = fread(buffer + size, 1, MAXLINELENGTH - size, file);
			if (n == 0) break;
			buffer[size + n] = 0;
			for (p=pos; (*p != 0) && (*p != '\n'); p++);
			if (*p == 0) { fclose(file); return 3; }
		}

		// ignore line with white space only
		while (*pos == ' ' || *pos == '\t') pos++;
		if (*pos == '\n')
		{
			pos++;
			continue;
		}

		// parse label
		VALUE l = strtod(pos, &end);
		if (end == pos) { fclose(file); return 2; }
		pos = end;
		if (mode == 1) data->label[point] = ((l == label[0]) ? -1.0 : +1.0);
		else data->label[point] = l;

		// parse features
		while (1)
		{
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos == '\n')
			{
				pos++;
				break;
			}

			INDEX index = strtoll(pos, &end, 10);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			if (! has_zero) index--;
			while (*pos == ' ' || *pos == '\t') pos++;
			if (*pos != ':') { fclose(file); return 2; }
			else pos++;
			while (*pos == ' ' || *pos == '\t') pos++;
			VALUE value = strtod(pos, &end);
			if (end == pos) { fclose(file); return 2; }
			pos = end;
			data->data[index][feature[index]].index = point;
			data->data[index][feature[index]].value = value;
			feature[index]++;
		}

		point++;
	}

	fclose(file);
	free(feature);
	printf(" done.\n");

	printf("  labels:     %s\n", (mode == 1) ? "binary classification" : "regression");
	printf("  points:     %u\n", (unsigned int)num_points);
	printf("  features:   %u\n", (unsigned int)num_features);
	printf("  entries:    %lu\n", num_entries);

	return 0;
}


////////////////////////////////////////////////////////////
// LASSO training
//


// return 0 on success and an error code otherwise:
// 1: out of memory
int train_cd(struct Data* data, struct Hypothesis* hyp, VALUE lambda, VALUE accuracy)
{
	printf("training with cyclic CD [lambda=%g] ...", lambda); fflush(stdout);

	INDEX num = data->num;

	// allocate memory
	VALUE* alpha = (VALUE*)(void*)calloc(data->num, sizeof(VALUE));
	VALUE* diag = (VALUE*)(void*)malloc(data->num * sizeof(VALUE));
	VALUE* w = (VALUE*)(void*)malloc(data->ell * sizeof(VALUE));
	if (alpha == NULL || diag == NULL || w == NULL)
	{
		free(alpha);
		free(diag);
		free(w);
		return 1;
	}

	// pre-calculate diagonal matrix entries (feature-wise squared norms)
	for (INDEX i=0; i<num; i++)
	{
		VALUE sum = 0.0;
		for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) sum += e->value * e->value;
		diag[i] = sum;
	}

	// initialize state vector w
	for (INDEX i=0; i<data->ell; i++) w[i] = -data->label[i];

	// main optimization loop
	size_t iter = 0;
	size_t steps = 0;
	size_t ops = 0;
	while (1)
	{
		VALUE maxvio = 0.0;

		steps += num;
		for (INDEX i=0; i<num; i++)
		{
			VALUE a = alpha[i];
			VALUE d = diag[i];

			ops += (data->data[i+1] - data->data[i]);

			// compute "gradient component" <w, X_i>
			VALUE grad = 0.0;
			for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) grad += w[e->index] * e->value;

			// compute optimal coordinate descent step and corresponding gain
			VALUE vio = 0.0;
			VALUE delta = 0.0;
			if (a == 0.0)
			{
				if (grad > lambda)
				{
					vio = grad - lambda;
					delta = -vio / d;
				}
				else if (grad < -lambda)
				{
					vio = -grad - lambda;
					delta = vio / d;
				}
			}
			else if (a > 0.0)
			{
				grad += lambda;
				vio = fabs(grad);
				delta = -grad / d;
				if (delta < -a)
				{
					delta = -a;
					VALUE g0 = grad - a * d - 2.0 * lambda;
					if (g0 > 0.0)
					{
						VALUE dd = -g0 / d;
						delta += dd;
					}
				}
			}
			else
			{
				grad -= lambda;
				vio = fabs(grad);
				delta = -grad / d;
				if (delta > -a)
				{
					delta = -a;
					VALUE g0 = grad - a * d + 2.0 * lambda;
					if (g0 < 0.0)
					{
						VALUE dd = -g0 / d;
						delta += dd;
					}
				}
			}

			// update state
			if (vio > maxvio) maxvio = vio;
			if (delta != 0.0)
			{
				alpha[i] += delta;
				for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) w[e->index] += delta * e->value;
			}
		}
		iter++;
		printf(".");
		fflush(stdout);

		if (maxvio <= accuracy) break;
	}

	printf(" done.\n");

	// compute objective value and count non-zero features
	INDEX nnz = 0;
	VALUE obj = 0.0;
	for (INDEX i=0; i<num; i++)
	{
		if (alpha[i] != 0.0)
		{
			nnz++;
			obj += fabs(alpha[i]);
		}
	}
	obj *= lambda;
	for (INDEX i=0; i<data->ell; i++) obj += w[i] * w[i];

	// output statistics
	printf("  update steps:        %lu\n", steps);
	printf("  'operations':        %lu\n", ops);
	printf("  objective value:     %f\n", obj);
	printf("  non-zero features:   %u\n", (unsigned int)nnz);

	// initialize hypothesis
	hyp->num = num;
	hyp->weight = (struct Entry*)(void*)malloc((nnz + 1) * sizeof(struct Entry));
	if (hyp->weight == NULL)
	{
		free(alpha);
		free(diag);
		free(w);
		return 1;
	}
	for (INDEX j=0, i=0; i<num; i++)
	{
		if (alpha[i] != 0.0)
		{
			hyp->weight[j].index = i;
			hyp->weight[j].value = alpha[i];
		}
	}
	hyp->weight[nnz].index = endOfVector;
	hyp->bias = 0.0;

	// clean up
	free(alpha);
	free(diag);
	free(w);

	return 0;
}


// strategy constants
#define CHANGE_RATE 0.2
#define PREF_MIN 0.05
#define PREF_MAX 20.0


// return 0 on success and an error code otherwise:
// 1: out of memory
// 2: too many features (get a better random number generator)
int train_acf(struct Data* data, struct Hypothesis* hyp, VALUE lambda, VALUE accuracy)
{
	printf("training with ACF-CD [lambda=%g] ...", lambda); fflush(stdout);

	INDEX num = data->num;
	if (num > (INDEX)RAND_MAX) return 2;

	// allocate memory
	VALUE* alpha = (VALUE*)(void*)calloc(data->num, sizeof(VALUE));
	VALUE* diag = (VALUE*)(void*)malloc(data->num * sizeof(VALUE));
	VALUE* w = (VALUE*)(void*)malloc(data->ell * sizeof(VALUE));
	INDEX* index = (INDEX*)(void*)malloc(2 * num * sizeof(INDEX));
	VALUE* pref = (VALUE*)(void*)malloc(num * sizeof(VALUE));
	VALUE* acc = (VALUE*)(void*)malloc(num * sizeof(VALUE));
	if (alpha == NULL || diag == NULL || w == NULL || index == NULL || pref == NULL || acc == NULL)
	{
		free(alpha);
		free(diag);
		free(w);
		free(index);
		free(pref);
		free(acc);
		return 1;
	}

	// pre-calculate diagonal matrix entries (feature-wise squared norms)
	for (INDEX i=0; i<num; i++)
	{
		VALUE sum = 0.0;
		for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) sum += e->value * e->value;
		diag[i] = sum;
	}

	// initialize state vector w
	for (INDEX i=0; i<data->ell; i++) w[i] = -data->label[i];

	// prepare preferences for scheduling
	for (INDEX i=0; i<num; i++) pref[i] = 1.0;
	VALUE prefsum = (VALUE)num;

	// prepare accumulators
	for (INDEX i=0; i<num; i++) acc[i] = 0.5;

	// prepare performance monitoring for self-adaptation
	const VALUE gain_learning_rate = 1.0 / num;
	VALUE average_gain = 0.0;
	int canstop = 1;

	// main optimization loop
	size_t iter = 0;
	size_t steps = 0;
	size_t ops = 0;
	while (1)
	{
		VALUE maxvio = 0.0;

		// define schedule
		INDEX slen = 0;
		VALUE q = num / prefsum;
		for (INDEX i=0; i<num; i++)
		{
			VALUE a = acc[i] + q * pref[i];
			int n = (int)floor(a);
			for (int j=0; j<n; j++)
			{
				index[slen] = i;
				slen++;
			}
			acc[i] = a - n;
		}
		for (INDEX i=0; i<num; i++)
		{
			INDEX* ptr = index + (rand() % num);
			INDEX tmp = index[i];
			index[i] = *ptr;
			*ptr = tmp;
		}

		steps += slen;
		for (INDEX s=0; s<slen; s++)
		{
			INDEX i = index[s];

			VALUE a = alpha[i];
			VALUE d = diag[i];

			ops += (data->data[i+1] - data->data[i]);

			// compute $\langle w, X_i\rangle$
			VALUE grad = 0.0;
			for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) grad += w[e->index] * e->value;

			// compute optimal coordinate descent step and corresponding gain
			VALUE vio = 0.0;
			VALUE gain = 0.0;
			VALUE delta = 0.0;
			if (a == 0.0)
			{
				if (grad > lambda)
				{
					vio = grad - lambda;
					delta = -vio / d;
					gain = 0.5 * d * delta * delta;
				}
				else if (grad < -lambda)
				{
					vio = -grad - lambda;
					delta = vio / d;
					gain = 0.5 * d * delta * delta;
				}
			}
			else if (a > 0.0)
			{
				grad += lambda;
				vio = fabs(grad);
				delta = -grad / d;
				if (delta < -a)
				{
					delta = -a;
					gain = delta * (grad - 0.5 * d * delta);
					VALUE g0 = grad - a * d - 2.0 * lambda;
					if (g0 > 0.0)
					{
						VALUE dd = -g0 / d;
						gain = dd * (grad - 0.5 * d * dd);
						delta += dd;
					}
				}
				else gain = 0.5 * d * delta * delta;
			}
			else
			{
				grad -= lambda;
				vio = fabs(grad);
				delta = -grad / d;
				if (delta > -a)
				{
					delta = -a;
					gain = delta * (grad - 0.5 * d * delta);
					VALUE g0 = grad - a * d + 2.0 * lambda;
					if (g0 < 0.0)
					{
						VALUE dd = -g0 / d;
						gain = dd * (grad - 0.5 * d * dd);
						delta += dd;
					}
				}
				else gain = 0.5 * d * delta * delta;
			}

			// update state
			if (vio > maxvio) maxvio = vio;
			if (delta != 0.0)
			{
				alpha[i] += delta;
				for (struct Entry* e = data->data[i]; e != data->data[i+1]; e++) w[e->index] += delta * e->value;
			}

			// update gain-based preferences
			{
				if (iter == 0) average_gain += gain / (VALUE)num;
				else
				{
					VALUE oldpref = pref[i];
					VALUE change = CHANGE_RATE * (gain / average_gain - 1.0);
					VALUE newpref = oldpref * exp(change);
					if (newpref < PREF_MIN) newpref = PREF_MIN;
					else if (newpref > PREF_MAX) newpref = PREF_MAX;
					prefsum += newpref - oldpref;
					pref[i] = newpref;
					average_gain = (1.0 - gain_learning_rate) * average_gain + gain_learning_rate * gain;
				}
			}
		}
		iter++;

		if (maxvio <= accuracy)
		{
			if (canstop) break;
			else
			{
				// prepare full sweep for a reliable check of the stopping criterion
				canstop = 1;
				for (INDEX i=0; i<num; i++) acc[i] = 0.5;
				for (INDEX i=0; i<num; i++) pref[i] = 1.0;
				prefsum = (VALUE)num;
				printf("*");
				fflush(stdout);
			}
		}
		else
		{
			canstop = 0;
			printf(".");
			fflush(stdout);
		}
	}

	printf(" done.\n");

	// compute objective value and count non-zero features
	INDEX nnz = 0;
	VALUE obj = 0.0;
	for (INDEX i=0; i<num; i++)
	{
		if (alpha[i] != 0.0)
		{
			nnz++;
			obj += fabs(alpha[i]);
		}
	}
	obj *= lambda;
	for (INDEX i=0; i<data->ell; i++) obj += w[i] * w[i];

	// output statistics
	printf("  update steps:        %lu\n", steps);
	printf("  'operations':        %lu\n", ops);
	printf("  objective value:     %f\n", obj);
	printf("  non-zero features:   %u\n", (unsigned int)nnz);

	// initialize hypothesis
	hyp->num = num;
	hyp->weight = (struct Entry*)(void*)malloc((nnz + 1) * sizeof(struct Entry));
	if (hyp->weight == NULL)
	{
		free(alpha);
		free(diag);
		free(w);
		free(index);
		free(pref);
		free(acc);
		return 1;
	}
	for (INDEX j=0, i=0; i<num; i++)
	{
		if (alpha[i] != 0.0)
		{
			hyp->weight[j].index = i;
			hyp->weight[j].value = alpha[i];
		}
	}
	hyp->weight[nnz].index = endOfVector;
	hyp->bias = 0.0;

	// clean up
	free(alpha);
	free(diag);
	free(w);
	free(index);
	free(pref);
	free(acc);

	return 0;
}


////////////////////////////////////////////////////////////
// help text
//

void help()
{
	printf(
			"\n"
			"LASSO regression\n"
			"usage: lasso <filename> <lambda> [<accuracy>] [<strategy>]\n"
			"   <filename>       (sparse) training data; data matrix A and labels y\n"
			"   <lambda>         regularization parameter \\lambda >= 0\n"
			"   <accuracy>       positive accuracy parameter; maximal deviation of\n"
			"                    regression outputs from their optimal values on the\n"
			"                    training data (default: 0.001)\n"
			"   <strategy>       0: cyclic coordinate descent (baseline)\n"
			"                    1: adaptive coordinate frequencies (ACF-CD)\n"
			"The code solves the problem\n"
			"   min \\lambda \\|w\\|_1 + \\frac{1}{2} \\|X w - y\\|_2^2\n"
			"where X is the data matrix (columns are instances, rows are features)\n"
			"and y is the vector of target values (labels).\n"
			"The program computes the (usually sparse) coefficient vector w.\n"
			"\n"
			);
	exit(0);
}


////////////////////////////////////////////////////////////
// entry point
//

int main(int argc, char** argv)
{
	if (argc < 3) help();
	int result;

	char* filename = argv[1];
	VALUE lambda = strtod(argv[2], NULL);
	if (lambda < 0.0) help();
	VALUE accuracy = 0.001;
	if (argc > 3) accuracy = strtod(argv[3], NULL);
	if (accuracy <= 0.0) help();
	int strategy = 0;
	if (argc > 4) strategy = strtol(argv[4], NULL, 10);
	if (strategy < 0 || strategy > 1) help();

	struct Data data;
	result = load(filename, &data);
	if (result != 0)
	{
		if (result == 1) printf("file open failed\n");
		if (result == 2) printf("file format violation\n");
		if (result == 3) printf("line too long\n");
		if (result == 4) printf("out of memory\n");
		if (result == 5) printf("unexpected label value\n");
		exit(1);
	}

	struct Hypothesis hyp;
	if (strategy == 0) result = train_cd(&data, &hyp, lambda, accuracy);
	if (strategy == 1) result = train_acf(&data, &hyp, lambda, accuracy);
	if (result != 0)
	{
		if (result == 1) printf("out of memory\n");
		if (result == 2) printf("too many features (better random number generator needed)\n");
		exit(1);
	}

	return 0;
}
